package com.yanzhu.flowable.service.impl;

import com.github.pagehelper.PageHelper;
import com.github.pagehelper.PageInfo;
import com.yanzhu.common.core.constant.HttpStatus;
import com.yanzhu.common.core.constant.SecurityConstants;
import com.yanzhu.common.core.domain.R;
import com.yanzhu.common.core.enums.FlowableEnums;
import com.yanzhu.common.core.exception.ServiceException;
import com.yanzhu.common.core.text.Convert;
import com.yanzhu.common.core.utils.StringUtils;
import com.yanzhu.common.core.web.domain.AjaxResult;
import com.yanzhu.common.core.web.page.TableDataInfo;
import com.yanzhu.common.redis.service.RedisService;
import com.yanzhu.common.security.utils.SecurityUtils;
import com.yanzhu.flowable.common.constant.ProcessConstants;
import com.yanzhu.flowable.common.enums.FlowComment;
import com.yanzhu.flowable.domain.FlowProcDefDto;
import com.yanzhu.flowable.domain.SysExpression;
import com.yanzhu.flowable.domain.vo.FlowDeptVo;
import com.yanzhu.flowable.domain.vo.FlowQueryVo;
import com.yanzhu.flowable.factory.FlowServiceFactory;
import com.yanzhu.flowable.mapper.FlowDeployMapper;
import com.yanzhu.flowable.service.IFlowDefinitionService;
import com.yanzhu.flowable.service.ISysDeployFormService;
import com.yanzhu.system.api.RemoteFlowService;
import com.yanzhu.system.api.domain.SysUser;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.io.IOUtils;
import org.flowable.bpmn.model.BpmnModel;
import org.flowable.engine.repository.Deployment;
import org.flowable.engine.repository.ProcessDefinition;
import org.flowable.engine.repository.ProcessDefinitionQuery;
import org.flowable.engine.runtime.ProcessInstance;
import org.flowable.image.impl.DefaultProcessDiagramGenerator;
import org.flowable.task.api.Task;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import javax.annotation.Resource;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.util.*;

/**
 * 流程定义
 *
 * @author Tony
 * @date 2021-04-03
 */
@Service
@Slf4j
public class FlowDefinitionServiceImpl extends FlowServiceFactory implements IFlowDefinitionService {

    @Autowired
    private RedisService redisService;

    @Resource
    private FlowDeployMapper flowDeployMapper;

    @Autowired
    private RemoteFlowService remoteFlowService;

    @Resource
    private ISysDeployFormService sysDeployFormService;

    private static final String BPMN_FILE_SUFFIX = ".bpmn";

    @Override
    public boolean exist(String processDefinitionKey) {
        ProcessDefinitionQuery processDefinitionQuery
                = repositoryService.createProcessDefinitionQuery().processDefinitionKey(processDefinitionKey);
        long count = processDefinitionQuery.count();
        return count > 0 ? true : false;
    }


    /**
     * 流程定义列表
     *
     * @param flowQueryVo  流程查询
     * @return 流程定义分页列表数据
     */
    @Override
    public TableDataInfo list(FlowQueryVo flowQueryVo) {
        TableDataInfo rspData = new TableDataInfo();
        PageHelper.startPage(flowQueryVo.getPageNum(), flowQueryVo.getPageSize());
        List<FlowProcDefDto> list = flowDeployMapper.selectDeployList(flowQueryVo);
        rspData.setCode(HttpStatus.SUCCESS);
        rspData.setMsg("查询成功");
        rspData.setRows(list);
        rspData.setTotal(new PageInfo(list).getTotal());
        return rspData;
    }


    /**
     * 导入流程文件
     *
     * 当每个key的流程第一次部署时，指定版本为1。对其后所有使用相同key的流程定义，
     * 部署时版本会在该key当前已部署的最高版本号基础上加1。key参数用于区分流程定义
     * @param name 流程名称
     * @param category 流程类型
     * @param prtype 流程定义类型
     * @param OwnerDeptId 流程单位
     * @param OwnerProjectId 流程项目
     * @param in 流程文件流
     */
    @Override
    @Transactional
    public void importFile(String name, String category, String prtype, String OwnerDeptId, String OwnerProjectId, InputStream in) {
        if(SecurityUtils.isAdmin(SecurityUtils.getUserId())){
            throw new ServiceException("请选择项目后导入流程文件...");
        }
        Deployment deploy = repositoryService.createDeployment().addInputStream(name + BPMN_FILE_SUFFIX, in).name(name).category(category).deploy();
        ProcessDefinition definition = repositoryService.createProcessDefinitionQuery().deploymentId(deploy.getId()).singleResult();
        repositoryService.setProcessDefinitionCategory(definition.getId(), category);
        /**新增流程单位关系**/
        FlowDeptVo flowDeptVo = new FlowDeptVo();
        flowDeptVo.setProcdefId(definition.getId());
        //超管不能导入流程文件
        if(SecurityUtils.isGSAdmin()){
            if(StringUtils.isNotEmpty(OwnerDeptId)){
                // OwnerDeptId转换Long失败...默认返回null
                flowDeptVo.setDeptId(Convert.toLong(OwnerDeptId));
            }
            if(StringUtils.isNotEmpty(OwnerDeptId)){
                // OwnerDeptId转换Long失败...默认返回null
                flowDeptVo.setProjId(Convert.toLong(OwnerProjectId));
            }
        }else{
            // 从用户信息中获取项目单位信息...
            flowDeptVo.setPrtype(prtype);
            flowDeptVo.setProjId(SecurityUtils.getLoginUser().getProjectId());
            flowDeptVo.setDeptId(SecurityUtils.getLoginUser().getProjectDeptId());
        }
        if(flowDeptVo.getDeptId()!=null){
            flowDeployMapper.insertActReProcdefDept(flowDeptVo);
        }
    }

    /**
     * 读取xml
     *
     * @param deployId
     * @return
     */
    @Override
    public AjaxResult readXml(String deployId) throws IOException {
        ProcessDefinition definition = repositoryService.createProcessDefinitionQuery().deploymentId(deployId).singleResult();
        InputStream inputStream = repositoryService.getResourceAsStream(definition.getDeploymentId(), definition.getResourceName());
        String result = IOUtils.toString(inputStream, StandardCharsets.UTF_8.name());
        return AjaxResult.success("", result);
    }

    /**
     * 读取xml
     *
     * @param deployId
     * @return
     */
    @Override
    public InputStream readImage(String deployId) {
        ProcessDefinition processDefinition = repositoryService.createProcessDefinitionQuery().deploymentId(deployId).singleResult();
        //获得图片流
        DefaultProcessDiagramGenerator diagramGenerator = new DefaultProcessDiagramGenerator();
        BpmnModel bpmnModel = repositoryService.getBpmnModel(processDefinition.getId());
        //输出为图片
        return diagramGenerator.generateDiagram(
                bpmnModel,
                "png",
                Collections.emptyList(),
                Collections.emptyList(),
                "宋体",
                "宋体",
                "宋体",
                null,
                1.0,
                false);

    }

    /**
     * 根据流程定义ID启动流程实例
     *
     * @param procDefId 流程模板ID
     * @param applyInfos 流程表单
     * @return
     */
    @Override
    public AjaxResult startProcessInstanceById(String procDefId, Map<String,Object> applyInfos) {
        try {
            /**保存流程表单**/
            Map<String, Object> procDefData = redisService.getCacheObject(FlowableEnums.FLOW_KEY_CACHE+procDefId);
            R<Long> res;
            if(StringUtils.isNotNull(procDefData)){
                String procType = Convert.toStr(procDefData.get("TYPE_"));
                if(StringUtils.isNotNull(procType)){
                    if(Objects.equals(procType,FlowableEnums.SUB_DEPTS.getCode())){
                        res = remoteFlowService.addSubdeptsEntity(applyInfos, SecurityConstants.INNER);
                    }else if(Objects.equals(procType,FlowableEnums.SUB_DEPTS.getCode())){
                        res = remoteFlowService.addSubdeptsGroupEntity(applyInfos, SecurityConstants.INNER);
                    }else if(Objects.equals(procType,FlowableEnums.SUB_DEPTS.getCode())){
                        res = remoteFlowService.addSubdeptsUsersEntity(applyInfos, SecurityConstants.INNER);
                    }else{
                        return AjaxResult.error("审批流程配置异常，请联系管理员配置");
                    }
                }else{
                    return AjaxResult.error("审批流程配置异常，请联系管理员配置");
                }
            }else{
                return AjaxResult.error("未配置审批流程信息，请联系管理员配置");
            }
            if(StringUtils.isNotNull(res.getData())){
                Map<String, Object> variables = new HashMap<>();
                ProcessDefinition processDefinition = repositoryService.createProcessDefinitionQuery().processDefinitionId(procDefId)
                        .latestVersion().singleResult();
                if (Objects.nonNull(processDefinition) && processDefinition.isSuspended()) {
                    return AjaxResult.error("流程已被挂起,请联系管理员激活流程");
                }
                SysUser sysUser = SecurityUtils.getLoginUser().getSysUser();
                String userId = Convert.toStr(sysUser.getUserId());
                // 设置流程发起人Id到流程中
                identityService.setAuthenticatedUserId(userId);
                variables.put(ProcessConstants.PROCESS_INITIATOR, userId);
                variables.put("businessKey", res.getData());
                ProcessInstance processInstance;
                processInstance = runtimeService.startProcessInstanceById(procDefId, Convert.toStr(res.getData()),variables);
                // 给第一步申请人节点设置任务执行人和意见
                Task task = taskService.createTaskQuery().processInstanceId(processInstance.getProcessInstanceId()).singleResult();
                if (Objects.nonNull(task)) {
                    taskService.addComment(task.getId(), processInstance.getProcessInstanceId(), FlowComment.NORMAL.getType(), sysUser.getNickName() + "发起流程申请");
                    taskService.setAssignee(task.getId(), Convert.toStr(userId));
                    taskService.complete(task.getId(), variables);
                }
                return AjaxResult.success("流程启动成功");
            }else{
                return AjaxResult.error("流程表单保存失败");
            }
        } catch (Exception e) {
            e.printStackTrace();
            return AjaxResult.error("流程启动错误");
        }
    }

    /**
     * 激活或挂起流程定义
     *
     * @param state    状态
     * @param deployId 流程部署ID
     */
    @Override
    public void updateState(Integer state, String deployId) {
        ProcessDefinition procDef = repositoryService.createProcessDefinitionQuery().deploymentId(deployId).singleResult();
        // 激活
        if (state == 1) {
            repositoryService.activateProcessDefinitionById(procDef.getId(), true, null);
        }
        // 挂起
        if (state == 2) {
            repositoryService.suspendProcessDefinitionById(procDef.getId(), true, null);
        }
    }


    /**
     * 删除流程定义
     *
     * @param deployId 流程部署ID act_ge_bytearray 表中 deployment_id值
     */
    @Override
    public void delete(String deployId) {
        // true 允许级联删除 ,不设置会导致数据库外键关联异常
        repositoryService.deleteDeployment(deployId, true);
    }

}
